# -*- coding: utf-8 -*-
"""

@author: Mike Tice and Tanner Mills

Redox profile model (daisy-chains models for aerobic respiration, iron
reduction, sulfate reduction, and methanogenesis)

Slow model assumes that carbonate precipitation and dissolution are slow
relative to respiration; carbonate mineral saturation index therefore varies
maximally and Ca cancentration is constant.
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.integrate import odeint
import pandas as pd

'''Definition of equilibrium constants. Select seawater or freshwater, 
or enter constants appropriate to your system.'''

system = 'seawater' # May be seawater or freshwater
mineral = 'calcite' # May be calcite or aragonite

if system == 'seawater':
    pKw = 14 
    pKc1 = 5.9
    pKc2 = 8.9
    pKs1 = 6.9
    pKs2 = 12.9
    pKclay1 = 6.6
    pKclay2 = 8.7
    pKb = -2.0 # Complexation constant for FeHCO3+
    pKc = -2.83 # Complexation constant for FeCO3
    pb1 = 9.4 # Beta for FeOH+
    pb2 = 20.5 # Beta for Fe(OH)2
    pb3 = 29.0 # Beta for Fe(OH)3-
    if mineral == 'aragonite': pKsp = 6.19 # Need to correct and add sources for all solubility products
    elif mineral == 'calcite': pKsp = 8.22
elif system == 'freshwater':
    pKw = 14
    pKc1 = 6.3
    pKc2 = 10.3
    pKs1 = 6.9
    pKs2 = 12.9
    pKb = -2.0 # Complexation constant for FeHCO3+
    pKc = -2.83 # Complexation constant for FeCO3
    pb1 = 9.4 # Beta for FeOH+
    pb2 = 20.5 # Beta for Fe(OH)2
    pb3 = 29.0 # Beta for Fe(OH)3-
    if mineral == 'aragonite': pKsp = 6.19
    elif mineral == 'calcite': pKsp = 8.22

'''Definition of functions returning chemical species and speciation fractions'''

def ten(x): return np.exp(x*np.log(10))

def H(pH): return ten(-pH)

def OH(pH): return ten(pH - pKw)

def aH2CO3(pH):
    alpha = 1/(1 + ten(pH - pKc1) + ten(2*pH - pKc1 - pKc2))
    return alpha

def aHCO3(pH):
    alpha = ten(pH - pKc1)*aH2CO3(pH)
    return alpha

def aCO3(pH):
    alpha = ten(pH - pKc2)*aHCO3(pH)
    return alpha
    
def aClayOH1(pH):
    alpha = ten(pKclay1-pH)/(ten(pKclay1-pH)+1)
    return alpha

def aClayOH2(pH):
    alpha = ten(pKclay2-pH)/(ten(pKclay2-pH)+1)
    return alpha

def aH2S(pH):
    alpha = 1/(1 + ten(pH - pKs1) + ten(2*pH - pKs1 - pKs2))
    return alpha

def aHS(pH):
    alpha = ten(pH - pKs1)*aH2S(pH)
    return alpha

def aS(pH):
    alpha = ten(pH - pKs2)*aHS(pH)
    return alpha

def Ab(pH):
    coeff = (ten(pKc1 - pH) + 1 + ten(pH - pKc2))*(ten(-pKb - 3*pH) + ten(-pKc - pKc2 - 2*pH))
    return coeff

def Bb(pH, CT, FeT):
    coeff = (FeT - CT)*(ten(-pKb - 3*pH) + ten(-pKc2 - pKc - 2*pH))
    coeff += (ten(-3*pH) + ten(-pb1 - 2*pH) + ten(-pb2 - pH) + ten(-pb3))*(ten(pKc1 - pH) + 1 + ten(pH - pKc2))
    return coeff

def Cb(pH, CT):
    coeff = -CT*(ten(-3*pH) + ten(-pb1 - 2*pH) + ten(-pb2 - pH) + ten(-pb3))
    return coeff

def HCO3(pH, CT, FeT=0):
    if FeT > 0:
        A, B, C = Ab(pH), Bb(pH, CT, FeT), Cb(pH, CT)
        conc = (-B + np.sqrt((B**2) - 4*A*C))/(2*A)
        return conc
    else:
        return CT*aHCO3(pH)

def H2CO3(pH, CT, FeT=0): return HCO3(pH, CT, FeT)*ten(pKc1 - pH)

def CO3(pH, CT, FeT=0): return HCO3(pH, CT, FeT)*ten(pH - pKc2)

def H2S(pH, ST): return ST*aH2S(pH)

def HS(pH, ST): return ST*aHS(pH)

def S(pH, ST): return ST*aS(pH)

def ClayO1(pH): return Clay01/(ten(pKclay1-pH)+1)

def ClayOH1(pH): return ClayO1(pH)*ten(pKclay1-pH)

def ClayO2(pH): return Clay02/(ten(pKclay2-pH)+1)

def ClayOH2(pH): return ClayO2(pH)*ten(pKclay2-pH)

def aFe(pH, CT, FeT):
    alpha = ten(-3*pH)/(ten(-3*pH) + (ten(-pKb - 3*pH) + ten(-pKc - pKc2 - 2*pH))*HCO3(pH, CT, FeT) + ten(-pb1 - 2*pH) + ten(-pb2 - pH) + ten(-pb3))
    return alpha

def aFeHCO3(pH, CT, FeT):
    alpha = aFe(pH, CT, FeT)*HCO3(pH, CT, FeT)*ten(-pKb)
    return alpha

def aFeCO3(pH, CT, FeT):
    alpha = aFe(pH, CT, FeT)*HCO3(pH, CT, FeT)*ten(pH - pKc - pKc2)
    return alpha

def aFeOH(pH, CT, FeT):
    alpha = aFe(pH, CT, FeT)*ten(pH - pb1)
    return alpha

def aFeOH2(pH, CT, FeT):
    alpha = aFe(pH, CT, FeT)*ten(2*pH - pb2)
    return alpha

def aFeOH3(pH, CT, FeT):
    alpha = aFe(pH, CT, FeT)*ten(3*pH - pb3)
    return alpha

def Beta(pH, CT, FeT):
    res = FeT*(-2*aFe(pH, CT, FeT) - aFeOH(pH, CT, FeT) + aFeOH3(pH, CT, FeT))*(aFeHCO3(pH, CT, FeT) + aFeCO3(pH, CT, FeT))
    res += FeT*aFeCO3(pH, CT, FeT)*(aFe(pH, CT, FeT) + aFeOH(pH, CT, FeT) + aFeOH2(pH, CT, FeT) + aFeOH3(pH, CT, FeT))
    res += -HCO3(pH, CT, FeT) - 2*CO3(pH, CT, FeT)
    return res

def Gamma(pH, Ca, CT, FeT):
    res = 2*Ca + 2*FeT*aFe(pH, CT, FeT)*(aFeCO3(pH, CT, FeT) + aFeOH(pH, CT, FeT) + 2*aFeOH2(pH, CT, FeT) + 3*aFeOH3(pH, CT, FeT))
    res += FeT*aFeHCO3(pH, CT, FeT)*(aFeCO3(pH, CT, FeT) + aFeOH(pH, CT, FeT) + 2*aFeOH2(pH, CT, FeT) + 3*aFeOH3(pH, CT, FeT))
    res += -FeT*aFeOH(pH, CT, FeT)*(1 - aFeCO3(pH, CT, FeT) -aFeOH(pH, CT, FeT) -2*aFeOH2(pH, CT, FeT) - 3*aFeOH3(pH, CT, FeT))
    res += FeT*aFeOH3(pH, CT, FeT)*(3 - aFeCO3(pH, CT, FeT) -aFeOH(pH, CT, FeT) -2*aFeOH2(pH, CT, FeT) - 3*aFeOH3(pH, CT, FeT))
    res += 2*CO3(pH, CT, FeT) + H(pH) + OH(pH)
    return res

def Delta(pH, Ca, CT, FeT):
    res = -2*Ca -2*FeT*aFe(pH, CT, FeT)*(aFeHCO3(pH, CT, FeT) + aFeCO3(pH, CT, FeT))
    res += FeT*aFeHCO3(pH, CT, FeT)*(1 - aFeHCO3(pH, CT, FeT) - aFeCO3(pH, CT, FeT))
    res += -FeT*(aFeOH(pH, CT, FeT) - aFeOH3(pH, CT, FeT))*(aFeHCO3(pH, CT, FeT) + aFeCO3(pH, CT, FeT))
    res += -HCO3(pH, CT, FeT) - 2*CO3(pH, CT, FeT)
    return res

def Epsilon(pH, Ca, CT, FeT):
    res = Ca + FeT*aFeCO3(pH, CT, FeT) + CO3(pH, CT, FeT) - H2CO3(pH, CT, FeT)
    res += -FeT*(aFeHCO3(pH, CT, FeT) + aFeCO3(pH, CT, FeT))*(aFeCO3(pH, CT, FeT) + aFeOH(pH, CT, FeT) + 2*aFeOH2(pH, CT, FeT) + 3*aFeOH3(pH, CT, FeT))
    return res

def Zeta(pH, Ca, CT, FeT):
    res = FeT*(aFeHCO3(pH, CT, FeT) + aFeCO3(pH, CT, FeT))*(1 - aFeHCO3(pH, CT, FeT) - aFeCO3(pH, CT, FeT))
    res += H2CO3(pH, CT, FeT) + HCO3(pH, CT, FeT) + CO3(pH, CT, FeT) + Ca
    return res

'''Definition of useful constants'''

M = 1
mM = 0.001*M
uM = 0.001*mM

'''Set boundary conditions'''

zone = 'aerobic' # Can be aerobic, ferruginous, sulfidic, or methanic
pH0 = 8 # Initial pH
DIC0 = 2.2*mM # Initial dissolved inorganic carbon
Clay01 = 10*mM #  Initial clay concentration, i.e., Clay0 = [clay:O-]+[clay:OH]
Clay02 = 10*mM
O2_consumed = 1*mM # Total oxygen respired at end of reaction
Fe_consumed = 2*mM # Total iron respired at end of reaction
SO4_consumed = 15*mM # Total sulfate respired at end of reaction
CH4_produced = 5*mM # Total methane produced at end of reaction
SI0 = 0.3 # Initial carbonate mineral saturation (for no precipitation model)
sp = 0 # Fraction of sulfate used to make pyrite

'''Calculating pH and SI'''

if (zone=='aerobic')==False :
    O2_consumed = 0*mM
    if (zone=='ferruginous')==False : Fe_consumed = 0*mM

N = 501
Ca0 = ten(-pKsp)/CO3(pH0, DIC0, 0)
f0 = [pH0, SI0, DIC0]
if zone=='aerobic' : x = np.linspace(0, O2_consumed, N)
else: x = [0]

def aer(y, x):
    pH, SI, CT = y
    F = -aHCO3(pH) - 2*aCO3(pH)
    beta = H(pH) + OH(pH) + HCO3(pH, CT)*(aH2CO3(pH) - aCO3(pH)) + 2*CO3(pH, CT)*(2*aH2CO3(pH) + aHCO3(pH))
    beta += ClayO1(pH)*aClayOH1(pH)
    beta += ClayO2(pH)*aClayOH2(pH)
    dpHdx = F/(beta*np.log(10))
    dSIdx = (2*aH2CO3(pH) + aHCO3(pH))*dpHdx + 1/(CT*np.log(10))
    return [dpHdx, dSIdx, 1]

if zone=='aerobic' :
    oxygen = odeint(aer, f0, x)
else: oxygen = f0

if zone=='aerobic' :
    pHslow, SI0, DICslow = oxygen[-1, 0], oxygen[-1, 1], oxygen[-1, 2]
else:
    pHslow, SI0, DICslow = pH0, SI0, DIC0
f0 = [pHslow, SI0, DICslow, 0]
if (zone=='aerobic') | (zone=='ferruginous') : y = np.linspace(0, Fe_consumed, N)
else: y = [0]

def fered(y, x):
    pH, SI, CT, FeT = y
    F = 2*aFe(pH, CT, FeT) + aFeHCO3(pH, CT, FeT) + aFeOH(pH, CT, FeT) - aFeOH3(pH, CT, FeT)
    F += Beta(pH, CT, FeT)*(0.25 - aFeHCO3(pH, CT, FeT) - aFeCO3(pH, CT, FeT))/(CT - FeT*(aFeHCO3(pH, CT, FeT) + aFeCO3(pH, CT, FeT)))
    beta = H(pH) + OH(pH) + 2*CO3(pH, CT, FeT)
    beta += 2*FeT*aFe(pH, CT, FeT)*(aFeCO3(pH, CT, FeT) + aFeOH(pH, CT, FeT) + 2*aFeOH2(pH, CT, FeT) + 3*aFeOH3(pH, CT, FeT))
    beta += FeT*aFeHCO3(pH, CT, FeT)*(aFeCO3(pH, CT, FeT) + aFeOH(pH, CT, FeT) + 2*aFeOH2(pH, CT, FeT))
    beta += FeT*aFeOH(pH, CT, FeT)*(aFe(pH, CT, FeT) + aFeOH(pH, CT, FeT) - aFeOH2(pH, CT, FeT) - 2*aFeOH3(pH, CT, FeT))
    beta += FeT*aFeOH3(pH, CT, FeT)*(3*aFe(pH, CT, FeT) + 3*aFeOH3(pH, CT, FeT) + 2*aFeCO3(pH, CT, FeT) + 2*aFeOH(pH, CT, FeT) + aFeOH3(pH, CT, FeT))
    beta += Beta(pH, CT, FeT)*(CO3(pH, CT, FeT) - H2CO3(pH, CT, FeT) + FeT*aFeCO3(pH, CT, FeT)*(aFe(pH, CT, FeT) + aFeHCO3(pH, CT, FeT) - aFeOH2(pH, CT, FeT) - 2*aFeOH3(pH, CT, FeT)) - FeT*aFeHCO3(pH, CT, FeT)*(aFeCO3(pH, CT, FeT) + aFeOH(pH, CT, FeT) + aFeOH2(pH, CT, FeT) + aFeOH3(pH, CT, FeT)))/(CT - FeT*(aFeHCO3(pH, CT, FeT) + aFeCO3(pH, CT, FeT)))
    beta += ClayO1(pH)*aClayOH1(pH)
    beta += ClayO2(pH)*aClayOH2(pH)
    dpHdx = F/(beta*np.log(10))
    dSIdx = 0
    return [dpHdx, dSIdx, 0.25, 1]


if (zone=='aerobic') | (zone=='ferruginous') :
    iron = odeint(fered, f0, y)
    for i in range(0, N): iron[i, 1] = np.log10(CO3(iron[i, 0], iron[i, 2], iron[i, 3])/CO3(iron[0, 0], iron[0, 2], iron[0, 3])) + SI0
else: iron = f0

if (zone=='aerobic') | (zone=='ferruginous') :
    pHslow, SI0, DICslow = iron[-1, 0], iron[-1, 1], iron[-1, 2]
else:
    pHslow, SI0, DICslow = pH0, SI0, DIC0

f0 = [pHslow, SI0, DICslow]
if (zone=='aerobic') | (zone=='ferruginous') | (zone=='sulfidic') : z = np.linspace(0, SO4_consumed, N)
else: z = [0]

def sulf(y, x):
    pH, SI, CT = y
    F = -(2 - 0.125*sp)*aHCO3(pH) - 2*(2 - 0.125*sp)*aCO3(pH) - (1 - sp)*aHS(pH) - 2*(1 - sp)*aS(pH) + 2
    beta = H(pH) + OH(pH) + HCO3(pH, CT)*(aH2CO3(pH) - aCO3(pH)) + 2*CO3(pH, CT)*(2*aH2CO3(pH) + aHCO3(pH)) + HS(pH, x)*(aH2S(pH) - aS(pH)) + 2*S(pH, x)*(2*aH2S(pH) + aHS(pH))
    beta += ClayO1(pH)*aClayOH1(pH)
    beta += ClayO2(pH)*aClayOH2(pH)
    dpHdx = F/(beta*np.log(10))
    dSIdx = (2*aH2CO3(pH) + aHCO3(pH))*dpHdx + (2 - 0.125*sp)/(CT*np.log(10))
    return [dpHdx, dSIdx, 2]


if (zone=='aerobic') | (zone=='ferruginous') | (zone=='sulfidic') :
    sulfate = odeint(sulf, f0, z)
else: sulfate = f0

if (zone=='aerobic') | (zone=='ferruginous') | (zone=='sulfidic') :
    pHslow, SI0, DICslow = sulfate[-1, 0], sulfate[-1, 1], sulfate[-1, 2]
else: 
    pHslow, SI0, DICslow = pH0, SI0, DIC0

f0 = [pHslow, SI0, DICslow]

if (zone=='aerobic') | (zone=='ferruginous') | (zone=='sulfidic') | (zone=='methanic') : m = np.linspace(0, CH4_produced, N)
else: m = [0]
def meth(y, x):
    pH, SI, CT = y
    F = aHCO3(pH) + 2*aCO3(pH)
    beta = H(pH) + OH(pH) + HCO3(pH, CT)*(aH2CO3(pH) - aCO3(pH)) 
    beta += 2*CO3(pH, CT)*(2*aH2CO3(pH) + aHCO3(pH))
    beta += ClayO1(pH)*aClayOH1(pH)
    beta += ClayO2(pH)*aClayOH2(pH)
    dpHdx = F/(beta*np.log(10))
    dSIdx = (2*aH2CO3(pH) + aHCO3(pH))*dpHdx + 1/(CT*np.log(10))
    return [dpHdx, dSIdx, -1]

methanic = odeint(meth, f0, m)

#np.savetxt('name_0.csv',oxygen,delimiter=',')        
#np.savetxt('name_Fe.csv',iron,delimiter=',')        
#np.savetxt('name_SO4.csv',sulfate,delimiter=',')
#np.savetxt('name_CH4.csv',methanic,delimiter=',')
#np.savetxt('namex.csv',x/mM,delimiter=',')
#np.savetxt('namey.csv',y/mM,delimiter=',')
#np.savetxt('namez.csv',z/mM,delimiter=',')
#np.savetxt('namem.csv',m/mM,delimiter=',')